/*
   Copyright (C) 2000  Ulric Eriksson <ulric@siag.nu>

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public
   License as published by the Free Software Foundation; either
   version 2 of the Licence, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.

   You should have received a copy of the GNU Library General Public
   License along with this library; if not, write to the Free
   Foundation, Inc., 59 Temple Place - Suite 330, Boston,
   MA 02111-1307, USA.
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <sys/stat.h>

#include <X11/IntrinsicP.h>
#include <X11/StringDefs.h>

#include "../common/cmalloc.h"
#include "../common/fonts.h"
#include "xcommon.h"
#include "xfonts.h"	/* from common */
#include <X11/xpm.h>

#include "RulerP.h"

static float floatOne = 1.0;

#define offset(field) XtOffsetOf(RulerRec, ruler.field)
static XtResource resources[] = {
	{
		XtNleftMargin,
		XtCLeftMargin,
		XtRInt,
		sizeof(int),
		offset(left_margin),
		XtRImmediate,
		(XtPointer)72
	}, {
		XtNrightMargin,
		XtCRightMargin,
		XtRInt,
		sizeof(int),
		offset(right_margin),
		XtRImmediate,
		(XtPointer)72
	}, {
		XtNpaperWidth,
		XtCPaperWidth,
		XtRInt,
		sizeof(int),
		offset(paper_width),
		XtRImmediate,
		(XtPointer)595
	}, {
		XtNrulerTopCol,
		XtCRulerTopCol,
		XtRInt,
		sizeof(int),
		offset(top_col),
		XtRImmediate,
		(XtPointer)0
	}, {
		XtNzoom,
		XtCZoom,
		XtRFloat,
		sizeof(float),
		offset(zoom),
		XtRFloat,
		(XtPointer)&floatOne
	}, {
		XtNfont,
		XtCFont,
		XtRFontStruct,
		sizeof(XFontStruct *),
		offset(font),
		XtRString,
		XtDefaultFont
	}, {
		XtNtabs,
		XtCTabs,
		XtRString,
		sizeof(String),
		offset(tabs),
		XtRImmediate,
		(XtPointer)0
	}, {
		XtNcallback,
		XtCCallback,
		XtRCallback,
		sizeof(XtPointer),
		offset(callbacks),
		XtRCallback,
		(XtPointer)NULL
	}
};
#undef offset

/* methods */
static void Realize(Widget w, XtValueMask *, XSetWindowAttributes *);
static void Redisplay(Widget, XEvent *, Region);
static Boolean SetValues(Widget, Widget, Widget, ArgList, Cardinal *);

/* actions */
static void Notify(Widget, XEvent *, String *, Cardinal *);
static void DeleteTab(Widget, XEvent *, String *, Cardinal *);
static void AddTab(Widget, XEvent *, String *, Cardinal *);

static XtActionsRec actions[] =
{
	{"notify", Notify},
	{"add_tab", AddTab},
	{"delete_tab", DeleteTab},
};

/* translations */
static char translations[] =
"Shift<Btn1Down>:		delete_tab() notify()\n"
":<Btn1Down>:			add_tab(l) notify()\n"
":<Btn2Down>:			add_tab(c) notify()\n"
":<Btn3Down>:			add_tab(r) notify()\n"
;

RulerClassRec rulerClassRec = {
  { /* core fields */
    /* superclass		*/	(WidgetClass) &widgetClassRec,
    /* class_name		*/	"Ruler",
    /* widget_size		*/	sizeof(RulerRec),
    /* class_initialize		*/	NULL,
    /* class_part_initialize	*/	NULL,
    /* class_inited		*/	FALSE,
    /* initialize		*/	NULL,
    /* initialize_hook		*/	NULL,
    /* realize			*/	Realize,
    /* actions			*/	actions,
    /* num_actions		*/	XtNumber(actions),
    /* resources		*/	resources,
    /* num_resources		*/	XtNumber(resources),
    /* xrm_class		*/	NULLQUARK,
    /* compress_motion		*/	TRUE,
    /* compress_exposure	*/	TRUE,
    /* compress_enterleave	*/	TRUE,
    /* visible_interest		*/	FALSE,
    /* destroy			*/	NULL,
    /* resize			*/	NULL,
    /* expose			*/	Redisplay,
    /* set_values		*/	SetValues,
    /* set_values_hook		*/	NULL,
    /* set_values_almost	*/	XtInheritSetValuesAlmost,
    /* get_values_hook		*/	NULL,
    /* accept_focus		*/	NULL,
    /* version			*/	XtVersion,
    /* callback_private		*/	NULL,
    /* tm_table			*/	translations,
    /* query_geometry		*/	XtInheritQueryGeometry,
    /* display_accelerator	*/	XtInheritDisplayAccelerator,
    /* extension		*/	NULL
  },
  { /* ruler fields */
    /* empty			*/	0
  }
};

WidgetClass rulerWidgetClass = (WidgetClass)&rulerClassRec;


tabtab RulerNextTab(Widget w, int t)
{
	RulerWidget rw = (RulerWidget)w;
	return next_tab(rw->ruler.tt, t);
}

static void Notify(Widget w, XEvent *event, String *params, Cardinal *n)
{
	RulerWidget rw = (RulerWidget)w;
	XtCallCallbackList(w, rw->ruler.callbacks, (XtPointer)rw->ruler.tabs);
}

static void DeleteTab(Widget w, XEvent *event, String *params, Cardinal *n)
{
	RulerWidget rw = (RulerWidget)w;
	float zoom = rw->ruler.zoom;
	int lm = rw->ruler.left_margin, tc = rw->ruler.top_col, ps = lm-tc+1;
	int x = (event->xbutton.x)/zoom-ps;
	int i;
	char *p = rw->ruler.tabs;
	tabtab *tt = rw->ruler.tt;
	/* no need to reallocate the tab string */
	for (i = 0; tt[i].j && tt[i].x < x; i++) {
		sprintf(p, "%c%d ", tt[i].j, tt[i].x);
		p += strlen(p);
	}
	if (tt[i].j) i++;
	while (tt[i].j) {
		sprintf(p, "%c%d ", tt[i].j, tt[i].x);
		p += strlen(p);
		i++;
	}
	cfree(rw->ruler.tt);
	rw->ruler.tt = get_tabs(rw->ruler.tabs);
	XClearWindow(XtDisplay(w), XtWindow(w));
	Redisplay(w, NULL, None);
}

static void AddTab(Widget w, XEvent *event, String *params, Cardinal *n)
{
	RulerWidget rw = (RulerWidget)w;
	float zoom = rw->ruler.zoom;
	int lm = rw->ruler.left_margin, tc = rw->ruler.top_col, ps = lm-tc+1;
	int x = (event->xbutton.x)/zoom-ps;
	int i;
	char *p;
	tabtab *tt = rw->ruler.tt;

	p = cmalloc(strlen(rw->ruler.tabs)+10);
	cfree(rw->ruler.tabs);
	rw->ruler.tabs = p;
	for (i = 0; tt[i].j && tt[i].x < x; i++) {
		sprintf(p, "%c%d ", tt[i].j, tt[i].x);
		p += strlen(p);
	}
	sprintf(p, "%c%d ", params[0][0], x);
	p += strlen(p);
	while (tt[i].j) {
		sprintf(p, "%c%d ", tt[i].j, tt[i].x);
		p += strlen(p);
		i++;
	}
	cfree(rw->ruler.tt);
	rw->ruler.tt = get_tabs(rw->ruler.tabs);
	XClearWindow(XtDisplay(w), XtWindow(w));
	Redisplay(w, NULL, None);
}

#define superclass (&coreClassRec)

static void Realize(Widget w, XtValueMask *v, XSetWindowAttributes *a)
{
	RulerWidget rw = (RulerWidget)w;
	unsigned long mask = 0;
	XGCValues values;
	(*superclass->core_class.realize)(w, v, a);
	values.font = rw->ruler.font->fid;
	mask = GCFont;
	rw->ruler.gc = XCreateGC(XtDisplay(w), XtWindow(w), mask, &values);
	if (rw->ruler.tabs == NULL) {
		rw->ruler.tabs = XtNewString("l36 l72 l108 l144 l180 l216 "
					     "l252 l288 l324 l360 l396 l432");
		rw->ruler.tt = get_tabs(rw->ruler.tabs);
	}
	rw->ruler.ctab = load_pixmap(w, "ctab.xpm");
	rw->ruler.ltab = load_pixmap(w, "ltab.xpm");
	rw->ruler.rtab = load_pixmap(w, "rtab.xpm");
}

static void Redisplay(Widget w, XEvent *event, Region r)
{
	RulerWidget aw = (RulerWidget)w;
	int h = aw->core.height;
	int w1 = aw->core.width;
	Display *dpy = XtDisplay(w);
	Screen *s = XtScreen(w);
	Window wi = XtWindow(w);
	int i, o, x;
	GC gc = aw->ruler.gc;
	unsigned long black = BlackPixelOfScreen(s);
	unsigned long white = WhitePixelOfScreen(s);
	float zoom = aw->ruler.zoom;
	int pw = aw->ruler.paper_width,
	    lm = aw->ruler.left_margin, rm = aw->ruler.right_margin,
	    tc = aw->ruler.top_col, ps = lm-tc+1, pe = pw-rm-tc, in;
	Pixmap pm;
	Pixmap scribble;

	scribble = XCreatePixmap(dpy, wi, w1, h, aw->core.depth);
	if (scribble == None) return;

	XSetForeground(dpy, gc, aw->core.background_pixel);
	XFillRectangle(dpy, scribble, gc, 0, 0, w1, h);

	XSetForeground(dpy, gc, white);

	XFillRectangle(dpy, scribble, gc, zoom*ps, 5, zoom*(pw-lm-rm), h-10);
	o = -1;

	XSetForeground(dpy, gc, black);

	/* draw tab stops */
	if (aw->ruler.tt == NULL) aw->ruler.tt = get_tabs(aw->ruler.tabs);
	for (i = 0; aw->ruler.tt[i].j; i++) {
		switch (aw->ruler.tt[i].j) {
		case 'c':	pm = aw->ruler.ctab; break;
		case 'r':	pm = aw->ruler.rtab; break;
		default:	pm = aw->ruler.ltab; break;
		}

		x = aw->ruler.tt[i].x;
		XCopyArea(dpy, pm, scribble, gc, 0, 0, 9, 5,
			zoom*(ps+x)-4+o, h-5);
	}

	/* draw half inch markers */
	for (i = ps+36; i < pe; i += 72) {
		XDrawLine(dpy, scribble, gc, zoom*i+o, 9, zoom*i+o, h-10);
	}

	/* draw inch numbers */
	for (i = ps+72, in = 1; i < pe; i += 72, in++) {
		char num[10];
		int nw;
		sprintf(num, "%d", in);
		nw = XTextWidth(aw->ruler.font, num, strlen(num));
		XDrawString(dpy, scribble, gc, zoom*i-nw/2, h-9, num, strlen(num));
	}
	XDrawLine(dpy, scribble, gc, zoom*-tc, 0, zoom*-tc, h-1);
	XDrawLine(dpy, scribble, gc, zoom*(pw-tc-1), 0, zoom*(pw-tc-1), h-1);
	XDrawLine(dpy, scribble, gc, 0, 0, 0, h-1);
	XDrawLine(dpy, scribble, gc, 0, 0, w1-1, 0);
	XDrawLine(dpy, scribble, gc, 0, h-1, w1-1, h-1);
	XDrawLine(dpy, scribble, gc, w1-1, 0, w1-1, h-1);

	XCopyArea(dpy, scribble, wi, gc, 0, 0, w1, h, 0, 0);
	XFreePixmap(dpy, scribble);
}

static Boolean SetValues(Widget current, Widget request, Widget new,
		ArgList args, Cardinal *nargs)
{
	RulerWidget rw1 = (RulerWidget)current;
	RulerWidget rw2 = (RulerWidget)new;
	Boolean do_redisplay =
			(rw1->ruler.tabs != rw2->ruler.tabs)
			|| strcmp(rw1->ruler.tabs, rw2->ruler.tabs)
			|| (rw1->ruler.zoom != rw2->ruler.zoom)
			|| (rw1->ruler.top_col != rw2->ruler.top_col)
			|| (rw1->ruler.left_margin != rw2->ruler.left_margin)
			|| (rw1->ruler.right_margin != rw2->ruler.right_margin)
			|| (rw1->ruler.paper_width != rw2->ruler.paper_width);
	if (rw1->ruler.tabs != rw2->ruler.tabs) {
		cfree(rw1->ruler.tabs);
		rw2->ruler.tabs = cstrdup(rw2->ruler.tabs);
		cfree(rw1->ruler.tt);
		rw2->ruler.tt = get_tabs(rw2->ruler.tabs);
	}

	return do_redisplay;
}

void RulerSetZoom(Widget w, float zoom)
{
	RulerWidget rw = (RulerWidget)w;

	if (zoom < .1) zoom = .1;
	if (zoom > 10) zoom = 10;
	if (zoom != rw->ruler.zoom) {
		rw->ruler.zoom = zoom;
		XClearWindow(XtDisplay(w), XtWindow(w));
		Redisplay(w, NULL, None);
	}
}

